package amqp

import (
	"fmt"
	"log"
	"net/url"

	"github.com/rabbitmq/amqp091-go"
)

// RabbitMQ 封装结构体
type RabbitMQ struct {
	conn    *amqp091.Connection
	channel *amqp091.Channel
}

// BuildRabbitMQURL 拼接RabbitMQ的URL
func BuildRabbitMQURL(username, password, hostname string, port int) string {
	u := &url.URL{
		Scheme: "amqp",
		User:   url.UserPassword(username, password),
		Host:   fmt.Sprintf("%s:%d", hostname, port),
		Path:   "/",
	}
	return u.String()
}

// NewRabbitMQ 实例化RabbitMQ连接
func NewRabbitMQ(username, password, hostname string, port int) (*RabbitMQ, error) {
	url := BuildRabbitMQURL(username, password, hostname, port)
	conn, err := amqp091.Dial(url)
	if err != nil {
		return nil, err
	}

	ch, err := conn.Channel()
	if err != nil {
		conn.Close()
		return nil, err
	}

	return &RabbitMQ{
		conn:    conn,
		channel: ch,
	}, nil
}

// Publish 发布消息
func (r *RabbitMQ) Publish(exchange, exchangeType, queueName, routingKey, body string) error {
	// 声明交换机
	err := r.channel.ExchangeDeclare(
		exchange,     // 交换机名称
		exchangeType, // 交换机类型
		true,         // 是否持久化
		false,        // 是否自动删除
		false,        // 是否为内部交换机
		false,        // 是否阻塞
		nil,          // 其他属性
	)
	if err != nil {
		return err
	}

	// 声明队列
	_, err = r.channel.QueueDeclare(
		queueName, // 队列名称
		true,      // 消息是否持久化
		false,     // 是否自动删除
		false,     // 是否为独占队列
		false,     // 是否阻塞
		nil,       // 其他属性
	)
	if err != nil {
		return err
	}

	// 绑定队列到交换机
	err = r.channel.QueueBind(
		queueName,  // 队列名称
		routingKey, // 路由键
		exchange,   // 交换机名称
		false,
		nil,
	)
	if err != nil {
		return err
	}

	err = r.channel.Publish(
		exchange,   // 交换机
		routingKey, // 路由键
		false,      // 是否返回消息
		false,      // 是否立即发送
		amqp091.Publishing{
			ContentType: "text/plain",
			Body:        []byte(body),
		})
	if err != nil {
		return err
	}

	log.Printf(" [x] Sent %s", body)
	return nil
}

// Consume 消费消息
func (r *RabbitMQ) Consume(exchange, exchangeType, queueName, routingKey string, handler func(amqp091.Delivery)) error {
	// 声明交换机
	err := r.channel.ExchangeDeclare(
		exchange,     // 交换机名称
		exchangeType, // 交换机类型
		true,         // 是否持久化
		false,        // 是否自动删除
		false,        // 是否为内部交换机
		false,        // 是否阻塞
		nil,          // 其他属性
	)
	if err != nil {
		return err
	}

	// 声明队列
	_, err = r.channel.QueueDeclare(
		queueName, // 队列名称
		true,      // 消息是否持久化
		false,     // 是否自动删除
		false,     // 是否为独占队列
		false,     // 是否阻塞
		nil,       // 其他属性
	)
	if err != nil {
		return err
	}

	// 绑定队列到交换机
	err = r.channel.QueueBind(
		queueName,  // 队列名称
		routingKey, // 路由键
		exchange,   // 交换机名称
		false,
		nil,
	)
	if err != nil {
		return err
	}

	// 消费消息
	msgs, err := r.channel.Consume(
		queueName, // 队列名称
		"",        // 消费者标签
		true,      // 自动应答
		false,     // 是否独占
		false,     // 是否为本地消费者
		false,     // 是否阻塞
		nil,       // 其他属性
	)
	if err != nil {
		return err
	}

	go func() {
		for d := range msgs {
			handler(d)
		}
	}()

	return nil
}

// Close 关闭连接和通道
func (r *RabbitMQ) Close() {
	if r.channel != nil {
		r.channel.Close()
	}
	if r.conn != nil {
		r.conn.Close()
	}
}

func failOnError(err error, msg string) {
	if err != nil {
		log.Fatalf("%s: %s", msg, err)
	}
}
