package one_orm

import (
	"fmt"
	"github.com/bytedance/sonic"
	"github.com/couchbase/gocb/v2"
)

type TransactionManager struct {
	db *CouchbaseDB
}

func NewTransactionManager(db *CouchbaseDB) *TransactionManager {
	return &TransactionManager{
		db: db,
	}
}

func (tm *TransactionManager) RunInTransaction(txnFunc func(ctx *gocb.TransactionAttemptContext) error, options *gocb.TransactionOptions) error {
	_, err := tm.db.cluster.Transactions().Run(txnFunc, options)
	if err != nil {
		return fmt.Errorf("transaction failed: %w", err)
	}
	return nil
}

func (tm *TransactionManager) SingleUpsertInTransaction(ctx *gocb.TransactionAttemptContext, collection *gocb.Collection, id string, doc interface{}) error {
	return tm.executeQuery(ctx, collection, "UPSERT", id, doc)
}

func (tm *TransactionManager) SingleReplaceInTransaction(ctx *gocb.TransactionAttemptContext, collection *gocb.Collection, id string, doc interface{}) error {
	getResult, err := ctx.Get(collection, id)
	if err != nil {
		return fmt.Errorf("failed to get document: %w", err)
	}
	_, err = ctx.Replace(getResult, doc)
	if err != nil {
		return fmt.Errorf("failed to replace document: %w", err)
	}
	return nil
}

func (tm *TransactionManager) SingleDeleteInTransaction(ctx *gocb.TransactionAttemptContext, collection *gocb.Collection, id string) error {
	getResult, err := ctx.Get(collection, id)
	if err != nil {
		return fmt.Errorf("failed to get document: %w", err)
	}
	err = ctx.Remove(getResult)
	if err != nil {
		return fmt.Errorf("failed to delete document: %w", err)
	}
	return nil
}

func (tm *TransactionManager) BatchUpsertInTransaction(ctx *gocb.TransactionAttemptContext, collection *gocb.Collection, docs map[string]interface{}) error {
	return tm.batchExecute(ctx, collection, docs, tm.SingleUpsertInTransaction)
}

func (tm *TransactionManager) BatchUpdateInTransaction(ctx *gocb.TransactionAttemptContext, collection *gocb.Collection, docs map[string]map[string]interface{}) error {
	return tm.batchExecute(ctx, collection, docs, tm.SingleReplaceInTransaction)
}

func (tm *TransactionManager) BatchDeleteInTransaction(ctx *gocb.TransactionAttemptContext, collection *gocb.Collection, ids []string) error {
	for _, id := range ids {
		err := tm.SingleDeleteInTransaction(ctx, collection, id)
		if err != nil {
			return err
		}
	}
	return nil
}

func (tm *TransactionManager) executeQuery(ctx *gocb.TransactionAttemptContext, collection *gocb.Collection, queryType string, id string, doc interface{}) error {
	docValue, err := sonic.Marshal(doc)
	if err != nil {
		return fmt.Errorf("failed to marshal document: %w", err)
	}
	query := fmt.Sprintf("%s INTO `%s`.`%s`.`%s` (KEY, VALUE) VALUES (\"%s\", %s);",
		queryType, tm.db.bucket.Name(), tm.db.scope.Name(), collection.Name(), id, docValue)
	_, err = ctx.Query(query, nil)
	if err != nil {
		return fmt.Errorf("failed to execute %s query: %w", queryType, err)
	}
	return nil
}

func (tm *TransactionManager) batchExecute(ctx *gocb.TransactionAttemptContext, collection *gocb.Collection, docs interface{}, singleFunc func(ctx *gocb.TransactionAttemptContext, collection *gocb.Collection, id string, doc interface{}) error) error {
	switch v := docs.(type) {
	case map[string]interface{}:
		for id, doc := range v {
			err := singleFunc(ctx, collection, id, doc)
			if err != nil {
				return err
			}
		}
	case map[string]map[string]interface{}:
		for id, doc := range v {
			err := singleFunc(ctx, collection, id, doc)
			if err != nil {
				return err
			}
		}
	default:
		return fmt.Errorf("unsupported docs type: %T", v)
	}
	return nil
}
