package one_orm

import (
	"fmt"
	"github.com/couchbase/gocb/v2"
	"reflect"
	"strings"
	"time"
)

type StorageEngine string

const (
	Magma      StorageEngine = "Magma"
	CouchStore StorageEngine = "CouchStore"
)

type CouchbaseDB struct {
	cluster    *gocb.Cluster
	bucket     *gocb.Bucket
	scope      *gocb.Scope
	collection *gocb.Collection
	engine     StorageEngine
}

type QueryBuilder[T any] struct {
	db               *CouchbaseDB
	statement        *Statement
	params           map[string]interface{}
	positionalParams []interface{}
	paramCount       int
	countDest        *int
	countField       string
	rawQuery         string
	scanConsistency  gocb.QueryScanConsistency
}

type Statement struct {
	Model      interface{}
	Table      string
	Selects    []string
	WhereConds []string
	Joins      []string
	Limit      int
	Offset     int
	Order      string
	Count      int
	GroupBys   []string
}

// NewCouchbaseDB 创建一个新的 CouchbaseDB 实例
// 如果没有传递存储引擎参数，则使用默认的 Magma 存储引擎
func NewCouchbaseDB(cluster *gocb.Cluster, bucket *gocb.Bucket, scope *gocb.Scope, collection *gocb.Collection, engine ...StorageEngine) *CouchbaseDB {
	selectedEngine := Magma
	if len(engine) > 0 {
		selectedEngine = engine[0]
	}
	// 设置 gocb 的日志记录
	//if isLoginEnabled {
	//	gocb.SetLogger(gocb.DefaultStdioLogger())
	//
	//}

	return &CouchbaseDB{
		cluster:    cluster,
		bucket:     bucket,
		scope:      scope,
		collection: collection,
		engine:     selectedEngine,
	}
}

func (db *CouchbaseDB) GetBucket() *gocb.Bucket {
	return db.bucket
}

func (db *CouchbaseDB) GetScope() *gocb.Scope {
	return db.scope
}

func (db *CouchbaseDB) GetCollection() *gocb.Collection {
	return db.collection
}

func (db *CouchbaseDB) GetEngine() StorageEngine {
	return db.engine
}

func NewQueryBuilder[T any](db *CouchbaseDB) *QueryBuilder[T] {
	return &QueryBuilder[T]{
		db: db,
		statement: &Statement{
			Selects:    []string{"*"},
			WhereConds: []string{},
			Joins:      []string{},
		},
		params:           make(map[string]interface{}),
		positionalParams: []interface{}{},
		paramCount:       0,
	}
}

func (qb *QueryBuilder[T]) Model(value interface{}) *QueryBuilder[T] {
	qb.statement.Model = value
	return qb
}

func (qb *QueryBuilder[T]) Table(name string) *QueryBuilder[T] {
	qb.statement.Table = name
	return qb
}

func (qb *QueryBuilder[T]) Select(fields ...string) *QueryBuilder[T] {
	qb.statement.Selects = fields
	return qb
}

func (qb *QueryBuilder[T]) Where(condition string, args ...interface{}) *QueryBuilder[T] {
	for i := range args {
		qb.paramCount++
		if qb.db.GetEngine() == Magma {
			placeholder := fmt.Sprintf("$param%d", qb.paramCount)
			condition = strings.Replace(condition, "?", placeholder, 1)
			qb.params[placeholder] = args[i]
		} else {
			condition = strings.Replace(condition, "?", fmt.Sprintf("$%d", qb.paramCount), 1)
			qb.positionalParams = append(qb.positionalParams, args[i])
		}
	}
	qb.statement.WhereConds = append(qb.statement.WhereConds, condition)
	return qb
}

func (qb *QueryBuilder[T]) Join(joinType, table, onCondition string) *QueryBuilder[T] {
	joinClause := fmt.Sprintf("%s JOIN %s ON %s", joinType, table, onCondition)
	qb.statement.Joins = append(qb.statement.Joins, joinClause)
	return qb
}

func (qb *QueryBuilder[T]) Limit(limit int) *QueryBuilder[T] {
	qb.statement.Limit = limit
	return qb
}

func (qb *QueryBuilder[T]) Offset(offset int) *QueryBuilder[T] {
	qb.statement.Offset = offset
	return qb
}

func (qb *QueryBuilder[T]) Order(order string) *QueryBuilder[T] {
	qb.statement.Order = order
	return qb
}

func (qb *QueryBuilder[T]) ScanConsistency(consistency gocb.QueryScanConsistency) *QueryBuilder[T] {
	qb.scanConsistency = consistency
	return qb
}

func (qb *QueryBuilder[T]) GroupBy(fields ...string) *QueryBuilder[T] {
	qb.statement.GroupBys = fields
	return qb
}

func (qb *QueryBuilder[T]) Like(field, pattern string) *QueryBuilder[T] {
	qb.paramCount++
	if qb.db.GetEngine() == Magma {
		placeholder := fmt.Sprintf("$param%d", qb.paramCount)
		condition := fmt.Sprintf("%s LIKE %s", field, placeholder)
		qb.statement.WhereConds = append(qb.statement.WhereConds, condition)
		qb.params[placeholder] = pattern
	} else {
		condition := fmt.Sprintf("%s LIKE $%d", field, qb.paramCount)
		qb.statement.WhereConds = append(qb.statement.WhereConds, condition)
		qb.positionalParams = append(qb.positionalParams, pattern)
	}
	return qb
}

func (qb *QueryBuilder[T]) isCountQuery() bool {
	return len(qb.statement.Selects) == 1 && strings.Contains(qb.statement.Selects[0], "COUNT")
}

func (qb *QueryBuilder[T]) Find(dest interface{}) error {
	if qb.countDest != nil {
		countQB := qb.toCountQueryBuilder()
		var countResult []map[string]interface{}
		err := countQB.Find(&countResult)
		if err != nil {
			return err
		}
		if len(countResult) > 0 {
			*qb.countDest = int(countResult[0]["count"].(float64))
		}
	}

	start := time.Now()
	query := qb.buildQuery()

	var err error
	var rows *gocb.QueryResult
	queryOptions := &gocb.QueryOptions{
		ScanConsistency: qb.scanConsistency,
		Adhoc:           true,
	}
	if qb.db.GetEngine() == Magma {
		queryOptions.NamedParameters = qb.params
	} else {
		queryOptions.PositionalParameters = qb.positionalParams
	}
	rows, err = qb.db.cluster.Query(query, queryOptions)
	if err != nil {
		return err
	}
	defer rows.Close()

	destVal := reflect.ValueOf(dest)
	if destVal.Kind() != reflect.Ptr || destVal.Elem().Kind() != reflect.Slice {
		return fmt.Errorf("dest must be a pointer to a slice")
	}
	sliceVal := destVal.Elem()
	elemType := sliceVal.Type().Elem()

	for rows.Next() {
		elem := reflect.New(elemType).Elem()
		err := rows.Row(elem.Addr().Interface())
		if err != nil {
			return err
		}
		sliceVal.Set(reflect.Append(sliceVal, elem))
	}

	elapsed := time.Since(start)
	fmt.Printf("Query executed in %s\n", elapsed)
	return nil
}

func (qb *QueryBuilder[T]) First(dest interface{}) error {
	qb.statement.Limit = 1

	start := time.Now()
	query := qb.buildQuery()

	var err error
	var rows *gocb.QueryResult
	queryOptions := &gocb.QueryOptions{
		ScanConsistency: qb.scanConsistency,
		Adhoc:           true,
	}
	if qb.db.GetEngine() == Magma {
		queryOptions.NamedParameters = qb.params
	} else {
		queryOptions.PositionalParameters = qb.positionalParams
	}
	rows, err = qb.db.cluster.Query(query, queryOptions)
	if err != nil {
		return err
	}
	defer rows.Close()

	destVal := reflect.ValueOf(dest)
	if destVal.Kind() != reflect.Ptr || destVal.Elem().Kind() != reflect.Struct {
		return fmt.Errorf("dest must be a pointer to a struct")
	}

	if rows.Next() {
		err := rows.Row(dest)
		if err != nil {
			return err
		}
	}

	elapsed := time.Since(start)
	fmt.Printf("Query executed in %s\n", elapsed)
	return nil
}
func (qb *QueryBuilder[T]) Count(dest *int, field string) *QueryBuilder[T] {
	qb.countDest = dest
	qb.countField = field
	return qb
}

func (qb *QueryBuilder[T]) toCountQueryBuilder() *QueryBuilder[map[string]interface{}] {
	return &QueryBuilder[map[string]interface{}]{
		db: qb.db,
		statement: &Statement{
			Selects:    []string{"COUNT(*) AS count"},
			WhereConds: qb.statement.WhereConds,
			Joins:      qb.statement.Joins,
		},
		params:           qb.params,
		positionalParams: qb.positionalParams,
	}
}

func (qb *QueryBuilder[T]) Scan(dest *[]T) *QueryBuilder[T] {
	err := qb.Find(dest)
	if err != nil {
		fmt.Println("Error in Scan: ", err)
	}
	return qb
}

func (qb *QueryBuilder[T]) Or(condition string, args ...interface{}) *QueryBuilder[T] {
	for i := range args {
		qb.paramCount++
		if qb.db.GetEngine() == Magma {
			placeholder := fmt.Sprintf("$param%d", qb.paramCount)
			condition = strings.Replace(condition, "?", placeholder, 1)
			qb.params[placeholder] = args[i]
		} else {
			condition = strings.Replace(condition, "?", fmt.Sprintf("$%d", qb.paramCount), 1)
			qb.positionalParams = append(qb.positionalParams, args[i])
		}
	}
	if len(qb.statement.WhereConds) > 0 {
		qb.statement.WhereConds[len(qb.statement.WhereConds)-1] = fmt.Sprintf("(%s OR %s)", qb.statement.WhereConds[len(qb.statement.WhereConds)-1], condition)
	} else {
		qb.statement.WhereConds = append(qb.statement.WhereConds, condition)
	}
	return qb
}

func (qb *QueryBuilder[T]) buildQuery() string {
	var queryBuilder strings.Builder

	if qb.rawQuery != "" {
		return qb.rawQuery
	}

	bucketName := qb.db.bucket.Name()
	scopeName := qb.db.scope.Name()
	collectionName := qb.db.collection.Name()

	queryBuilder.WriteString(fmt.Sprintf("SELECT %s FROM `%s`.`%s`.`%s`", strings.Join(qb.statement.Selects, ", "), bucketName, scopeName, collectionName))

	if len(qb.statement.Joins) > 0 {
		queryBuilder.WriteString(" " + strings.Join(qb.statement.Joins, " "))
	}

	if len(qb.statement.WhereConds) > 0 {
		queryBuilder.WriteString(" WHERE " + strings.Join(qb.statement.WhereConds, " AND "))
	}

	if len(qb.statement.GroupBys) > 0 {
		queryBuilder.WriteString(" GROUP BY " + strings.Join(qb.statement.GroupBys, ", "))
	}

	if qb.statement.Order != "" && !qb.isCountQuery() {
		queryBuilder.WriteString(" ORDER BY " + qb.statement.Order)
	}

	if qb.statement.Limit > 0 && !qb.isCountQuery() {
		queryBuilder.WriteString(fmt.Sprintf(" LIMIT %d", qb.statement.Limit))
	}

	if qb.statement.Offset > 0 && !qb.isCountQuery() {
		queryBuilder.WriteString(fmt.Sprintf(" OFFSET %d", qb.statement.Offset))
	}

	query := queryBuilder.String()
	fmt.Println("Query: ", query)
	return query
}
func (qb *QueryBuilder[T]) RawQuery(query string, args ...interface{}) *QueryBuilder[T] {
	qb.rawQuery = query
	for i, arg := range args {
		if qb.db.GetEngine() == Magma {
			qb.params[fmt.Sprintf("$param%d", i+1)] = arg
		} else {
			qb.positionalParams = append(qb.positionalParams, arg)
		}
	}
	return qb
}

func (qb *QueryBuilder[T]) CountRecords() (int, error) {
	bucketName := qb.db.bucket.Name()
	scopeName := qb.db.scope.Name()
	collectionName := qb.db.collection.Name()

	query := fmt.Sprintf("SELECT COUNT(*) AS count FROM `%s`.`%s`.`%s`", bucketName, scopeName, collectionName)

	if len(qb.statement.WhereConds) > 0 {
		query += " WHERE " + strings.Join(qb.statement.WhereConds, " AND ")
	}

	fmt.Println("Count Query: ", query)

	queryOptions := &gocb.QueryOptions{
		ScanConsistency: qb.scanConsistency,
		Adhoc:           true,
	}

	if qb.db.GetEngine() == Magma {
		queryOptions.NamedParameters = qb.params
	} else {
		queryOptions.PositionalParameters = qb.positionalParams
	}

	rows, err := qb.db.cluster.Query(query, queryOptions)
	if err != nil {
		return 0, err
	}
	defer rows.Close()

	var countResult struct {
		Count int `json:"count"`
	}

	if rows.Next() {
		err := rows.Row(&countResult)
		if err != nil {
			return 0, err
		}
	}

	return countResult.Count, nil
}
