package qywechatbot

import (
	"bytes"
	"encoding/json"
	"fmt"
	"net/http"
)

// Bot 是一个封装了企业微信机器人的结构体
type Bot struct {
	webhookURL string
}

// NewBot 初始化一个新的Bot实例
func NewBot(key string) *Bot {
	webhookURL := fmt.Sprintf("https://qyapi.weixin.qq.com/cgi-bin/webhook/send?key=%s", key)
	return &Bot{webhookURL: webhookURL}
}

// SendTextMessage 发送文本消息到微信群
func (b *Bot) SendTextMessage(content string, mentionedList, mentionedMobileList *[]string) error {
	text := map[string]interface{}{
		"content": content,
	}

	if mentionedList != nil {
		text["mentioned_list"] = *mentionedList
	}

	if mentionedMobileList != nil {
		text["mentioned_mobile_list"] = *mentionedMobileList
	}

	message := map[string]interface{}{
		"msgtype": "text",
		"text":    text,
	}

	return b.sendMessage(message)
}

// SendMarkdownMessage 发送Markdown消息
func (b *Bot) SendMarkdownMessage(content string) error {
	message := map[string]interface{}{
		"msgtype": "markdown",
		"markdown": map[string]string{
			"content": content,
		},
	}
	return b.sendMessage(message)
}

// SendImageMessage 发送图片消息
func (b *Bot) SendImageMessage(base64, md5 string) error {
	message := map[string]interface{}{
		"msgtype": "image",
		"image": map[string]string{
			"base64": base64,
			"md5":    md5,
		},
	}
	return b.sendMessage(message)
}

// SendNewsMessage 发送图文消息
func (b *Bot) SendNewsMessage(articles []map[string]string) error {
	message := map[string]interface{}{
		"msgtype": "news",
		"news": map[string]interface{}{
			"articles": articles,
		},
	}
	return b.sendMessage(message)
}

// SendFileMessage 发送文件消息
func (b *Bot) SendFileMessage(mediaID string) error {
	message := map[string]interface{}{
		"msgtype": "file",
		"file": map[string]string{
			"media_id": mediaID,
		},
	}
	return b.sendMessage(message)
}

// SendVoiceMessage 发送语音消息
func (b *Bot) SendVoiceMessage(mediaID string) error {
	message := map[string]interface{}{
		"msgtype": "voice",
		"voice": map[string]string{
			"media_id": mediaID,
		},
	}
	return b.sendMessage(message)
}

// sendMessage 发送消息的内部方法
func (b *Bot) sendMessage(message map[string]interface{}) error {
	messageJSON, err := json.Marshal(message)
	if err != nil {
		return fmt.Errorf("error encoding JSON: %w", err)
	}

	req, err := http.NewRequest("POST", b.webhookURL, bytes.NewBuffer(messageJSON))
	if err != nil {
		return fmt.Errorf("error creating request: %w", err)
	}

	req.Header.Set("Content-Type", "application/json")

	client := &http.Client{}
	resp, err := client.Do(req)
	if err != nil {
		return fmt.Errorf("error sending request: %w", err)
	}
	defer resp.Body.Close()

	if resp.StatusCode != http.StatusOK {
		return fmt.Errorf("failed to send message. Status code: %d", resp.StatusCode)
	}

	return nil
}
