package redis

import (
	"context"
	"errors"
	"fmt"
	"github.com/redis/go-redis/v9"
	"time"
)

var ErrKeyNotFound = errors.New("key not found")

func NewRedisClient(host string, port int, pass string, db int) *redis.Client {
	return redis.NewClient(&redis.Options{
		Addr:     fmt.Sprintf("%s:%d", host, port),
		Password: pass,
		DB:       db,
	})
}

type (
	RedisMethod interface {
		MGet(ctx context.Context, email []string) ([]interface{}, error)
		MSet(ctx context.Context, values ...interface{}) error
		Expire(ctx context.Context, key string, expiration time.Duration) error
		Expires(ctx context.Context, key []string, expiration time.Duration) error
		HSet(ctx context.Context, key string, values ...interface{}) error
		HGet(ctx context.Context, key, field string) (string, error)
		HGetAll(ctx context.Context, key string) (map[string]string, error)
		HDel(ctx context.Context, key string, fields ...string) error
		HMSet(ctx context.Context, key string, fields ...interface{}) (bool, error)
		HKeys(ctx context.Context, key string) ([]string, error)
		HScan(ctx context.Context, key string, icursor uint64, match string, count int64) (keys []string, ocursor uint64, err error)
		RPush(ctx context.Context, key string, fields ...interface{}) (int64, error)
		LPop(ctx context.Context, key string) (string, error)
		Del(ctx context.Context, keys ...string) error
		SAdd(ctx context.Context, key string, members ...interface{}) error
		SRem(ctx context.Context, key string, members ...interface{}) error
		SMembers(ctx context.Context, key string) ([]string, error)
		Keys(ctx context.Context, pattern string) ([]string, error)
		ScanAndDelete(ctx context.Context, pattern string) error
		Set(ctx context.Context, key string, value interface{}, expiration time.Duration) error
		Get(ctx context.Context, key string) (string, error)
		Append(ctx context.Context, key string, value string) (int64, error)
		StrLen(ctx context.Context, key string) (int64, error)
		Incr(ctx context.Context, key string) (int64, error)
		Decr(ctx context.Context, key string) (int64, error)
		SDiff(ctx context.Context, keys ...string) ([]string, error)
		SInter(ctx context.Context, keys ...string) ([]string, error)
		SUnion(ctx context.Context, keys ...string) ([]string, error)
		SPop(ctx context.Context, key string) (string, error)
		SCard(ctx context.Context, key string) (int64, error)
		SIsMember(ctx context.Context, key string, member interface{}) (bool, error)
	}
	defaultRedisConn struct {
		client *redis.Client
	}
)

func NewRedisMethod(host string, port int, pass string, db int) RedisMethod {
	return &defaultRedisConn{
		client: NewRedisClient(host, port, pass, db),
	}
}

func (r *defaultRedisConn) MGet(ctx context.Context, keys []string) ([]interface{}, error) {
	return r.client.MGet(ctx, keys...).Result()
}

func (r *defaultRedisConn) MSet(ctx context.Context, values ...interface{}) error {
	return r.client.MSet(ctx, values...).Err()
}

func (r *defaultRedisConn) Expire(ctx context.Context, key string, expiration time.Duration) error {
	return r.client.Expire(ctx, key, expiration).Err()
}

func (r *defaultRedisConn) Expires(ctx context.Context, keys []string, expiration time.Duration) error {
	for _, key := range keys {
		err := r.client.Expire(ctx, key, expiration).Err()
		if err != nil {
			return err
		}
	}
	return nil
}

// HSet 设置 Hash 的字段和值
func (r *defaultRedisConn) HSet(ctx context.Context, key string, values ...interface{}) error {
	return r.client.HSet(ctx, key, values...).Err()
}

// HGet 获取 Hash 的字段值
func (r *defaultRedisConn) HGet(ctx context.Context, key, field string) (string, error) {
	return r.client.HGet(ctx, key, field).Result()
}

// HGetAll 获取 Hash 的所有字段和值
func (r *defaultRedisConn) HGetAll(ctx context.Context, key string) (map[string]string, error) {
	return r.client.HGetAll(ctx, key).Result()
}

// HDel 删除 Hash 的字段
func (r *defaultRedisConn) HDel(ctx context.Context, key string, fields ...string) error {
	return r.client.HDel(ctx, key, fields...).Err()
}

// HScan 迭代哈希表中的键值对
func (r *defaultRedisConn) HScan(ctx context.Context, key string, icursor uint64, match string, count int64) (keys []string, ocursor uint64, err error) {
	return r.client.HScan(ctx, key, icursor, match, count).Result()
}

// RPush 用于将一个或多个值插入到列表的尾部(最右边)
func (r *defaultRedisConn) RPush(ctx context.Context, key string, fields ...interface{}) (int64, error) {
	return r.client.RPush(ctx, key, fields...).Result()
}

// HMSet 用于同时将多个 field-value (字段-值)对设置到哈希表中
func (r *defaultRedisConn) HMSet(ctx context.Context, key string, fields ...interface{}) (bool, error) {
	return r.client.HMSet(ctx, key, fields...).Result()
}

// LPop 用于移除并返回列表的第一个元素
func (r *defaultRedisConn) LPop(ctx context.Context, key string) (string, error) {
	return r.client.LPop(ctx, key).Result()
}

// Del 删除已存在的键。不存在的 key 会被忽略
func (r *defaultRedisConn) Del(ctx context.Context, keys ...string) error {
	return r.client.Del(ctx, keys...).Err()
}

// SAdd 将一个或多个成员元素加入到集合中，已经存在于集合的成员元素将被忽略
func (r *defaultRedisConn) SAdd(ctx context.Context, key string, members ...interface{}) error {
	return r.client.SAdd(ctx, key, members...).Err()
}

// SRem 移除集合中的一个或多个成员元素，不存在的成员元素会被忽略
func (r *defaultRedisConn) SRem(ctx context.Context, key string, members ...interface{}) error {
	return r.client.SRem(ctx, key, members...).Err()
}

// SMembers 返回集合中的所有的成员。 不存在的集合 key 被视为空集合
func (r *defaultRedisConn) SMembers(ctx context.Context, key string) ([]string, error) {
	return r.client.SMembers(ctx, key).Result()
}

// Keys 查找所有符合给定模式 pattern 的 key
func (r *defaultRedisConn) Keys(ctx context.Context, pattern string) ([]string, error) {
	return r.client.Keys(ctx, pattern).Result()
}

// HKeys 获取 Hash 的所有字段
func (r *defaultRedisConn) HKeys(ctx context.Context, key string) ([]string, error) {
	return r.client.HKeys(ctx, key).Result()
}

func (r *defaultRedisConn) ScanAndDelete(ctx context.Context, pattern string) error {
	var cursor uint64
	for {
		// 使用 SCAN 命令扫描匹配的键
		keys, nextCursor, err := r.client.Scan(ctx, cursor, pattern, 100).Result()
		if err != nil {
			return err
		}

		// 删除匹配的键
		if len(keys) > 0 {
			if err := r.Del(ctx, keys...); err != nil {
				return err
			}
		}

		// 如果 cursor 为 0，表示扫描结束
		if nextCursor == 0 {
			break
		}
		cursor = nextCursor
	}
	return nil
}

func (r *defaultRedisConn) Set(ctx context.Context, key string, value interface{}, expiration time.Duration) error {
	return r.client.Set(ctx, key, value, expiration).Err()
}

func (r *defaultRedisConn) Get(ctx context.Context, key string) (string, error) {
	val, err := r.client.Get(ctx, key).Result()
	if err == redis.Nil {
		return "", ErrKeyNotFound
	}
	return val, err
}

func (r *defaultRedisConn) Append(ctx context.Context, key string, value string) (int64, error) {
	return r.client.Append(ctx, key, value).Result()
}

func (r *defaultRedisConn) StrLen(ctx context.Context, key string) (int64, error) {
	return r.client.StrLen(ctx, key).Result()
}

func (r *defaultRedisConn) Incr(ctx context.Context, key string) (int64, error) {
	return r.client.Incr(ctx, key).Result()
}

func (r *defaultRedisConn) Decr(ctx context.Context, key string) (int64, error) {
	return r.client.Decr(ctx, key).Result()
}

func (r *defaultRedisConn) SIsMember(ctx context.Context, key string, member interface{}) (bool, error) {
	return r.client.SIsMember(ctx, key, member).Result()
}

func (r *defaultRedisConn) SCard(ctx context.Context, key string) (int64, error) {
	return r.client.SCard(ctx, key).Result()
}

func (r *defaultRedisConn) SPop(ctx context.Context, key string) (string, error) {
	return r.client.SPop(ctx, key).Result()
}

func (r *defaultRedisConn) SUnion(ctx context.Context, keys ...string) ([]string, error) {
	return r.client.SUnion(ctx, keys...).Result()
}

func (r *defaultRedisConn) SInter(ctx context.Context, keys ...string) ([]string, error) {
	return r.client.SInter(ctx, keys...).Result()
}

func (r *defaultRedisConn) SDiff(ctx context.Context, keys ...string) ([]string, error) {
	return r.client.SDiff(ctx, keys...).Result()
}
