package urlshortener

import (
	"bytes"
	"encoding/json"
	"fmt"
	"net/http"
)

type ShortenRequest struct {
	FullUrl        string `json:"fullUrl"`
	ExpirationTime int64  `json:"expirationTime"`
}

type ShortenResponse struct {
	Code    int    `json:"code"`
	Message string `json:"message"`
	Data    struct {
		ShortKey       string `json:"shortKey"`
		FullUrl        string `json:"fullUrl"`
		ExpirationTime int64  `json:"expirationTime"`
	} `json:"data"`
}

type URLShortenerClient struct {
	BaseURL    string
	HttpClient *http.Client
}

func NewClient(baseURL string) *URLShortenerClient {
	return &URLShortenerClient{
		BaseURL:    baseURL,
		HttpClient: &http.Client{},
	}
}

func (c *URLShortenerClient) ShortenURL(fullUrl string, expirationTime int64) (*ShortenResponse, error) {
	requestData := ShortenRequest{
		FullUrl:        fullUrl,
		ExpirationTime: expirationTime,
	}

	jsonData, err := json.Marshal(requestData)
	if err != nil {
		return nil, fmt.Errorf("failed to marshal request data: %v", err)
	}

	request, err := http.NewRequest("POST", c.BaseURL+"/url", bytes.NewBuffer(jsonData))
	if err != nil {
		return nil, fmt.Errorf("failed to create request: %v", err)
	}

	request.Header.Set("Content-Type", "application/json")

	response, err := c.HttpClient.Do(request)
	if err != nil {
		return nil, fmt.Errorf("failed to send request: %v", err)
	}
	defer response.Body.Close()

	if response.StatusCode != http.StatusOK {
		return nil, fmt.Errorf("unexpected status code: %d", response.StatusCode)
	}

	var shortenResponse ShortenResponse
	if err := json.NewDecoder(response.Body).Decode(&shortenResponse); err != nil {
		return nil, fmt.Errorf("failed to decode response: %v", err)
	}

	if shortenResponse.Code != 200 {
		return nil, fmt.Errorf("error from server: %s", shortenResponse.Message)
	}

	return &shortenResponse, nil
}
